from ..repository import BaseRepository
from ..schema import ListFilter,FindUniqueByFieldInput, AddColumnField


class BaseService:
    def __init__(self, repository: BaseRepository):
        self.repository = repository

    async def create(self, data: dict) -> any:
        return await self.repository.create(data)
    
    async def get_all(self,find:ListFilter) -> list[any]:
        return await self.repository.get_all(find)
    
    async def get_by_id(self, id: str) -> any:
        return await self.repository.get_by_id(id)

    async def update(self, id: str, data: dict) -> any:
        return await self.repository.update(id, data)
    
    async def delete(self, id: str) -> any:
        return await self.repository.delete(id)

    async def get_unique_values(self, schema: FindUniqueByFieldInput) -> list[any]:
        return await self.repository.get_unique_values(schema)
    
    async def add_field(self, column: AddColumnField):
        return await self.repository.add_field(column)
    
    async def bulk_create(self, data: list[any]) -> list[any]:
        return await self.repository.bulk_create(data)



from typing import Any, Dict, Optional

from fastapi import HTTPException, status


class DuplicatedError(HTTPException):
    def __init__(
        self, detail: Any = None, headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(status.HTTP_409_CONFLICT, detail, headers)


class AuthError(HTTPException):
    def __init__(
        self, detail: Any = None, headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(status.HTTP_403_FORBIDDEN, detail, headers)


class NotFoundError(HTTPException):
    def __init__(
        self, detail: Any = None, headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(status.HTTP_404_NOT_FOUND, detail, headers)


class ValidationError(HTTPException):
    def __init__(
        self, detail: Any = None, headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(status.HTTP_422_UNPROCESSABLE_ENTITY, detail, headers)


class PermissionDeniedError(HTTPException):
    def __init__(
        self, detail: Any = None, headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(status.HTTP_403_FORBIDDEN, detail, headers)


class UnauthorizedError(HTTPException):
    def __init__(
        self, detail: Any = None, headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(status.HTTP_401_UNAUTHORIZED, detail, headers)

class BadRequestError(HTTPException):
    def __init__(
        self, detail: Any = None, headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(status.HTTP_400_BAD_REQUEST, detail, headers)

class MongoError(HTTPException):
    def __init__(
        self, detail: Any = None, headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(status.HTTP_500_INTERNAL_SERVER_ERROR, detail, headers)



from datetime import datetime
from typing import Type

from ..schema import Operator, LogicalOperator
from ..document import BaseDocument


logical_operator_map = {
    LogicalOperator.AND: "$and",
    LogicalOperator.OR: "$or",
}



def apply_condition(model:Type[BaseDocument],operator:Operator,field:str,value:any):
        field_type = model.model_fields[field].annotation
        if field_type == datetime:
            value = datetime.fromisoformat(value)
        if operator == Operator.EQ:
            return {field: {"$eq": value}}
        elif operator == Operator.NE:
            return {field: {"$ne": value}}
        elif operator == Operator.GT:
            return {field: {"$gt": value}}
        elif operator == Operator.GTE:
            return {field: {"$gte": value}}
        elif operator == Operator.LT:
            return {field: {"$lt": value}}
        elif operator == Operator.LTE:
            return {field: {"$lte": value}}

        # Handle 'IN' and 'NIN'
        elif operator == Operator.IN:
            return {field: {"$in": value}}
        elif operator == Operator.NIN:
            return {field: {"$nin": value}}

        # Handle 'LIKE' (wildcard match) and 'ILIKE' (case-insensitive match)
        elif operator == Operator.LIKE:
            return {field: {"$regex": f".*{value}.*"}}  # Case-sensitive LIKE
        elif operator == Operator.ILIKE:
            return {field: {"$regex": f".*{value}.*", "$options": "i"}}  # Case-insensitive LIKE

        # Handle 'BETWEEN'
        elif operator == Operator.BETWEEN:
            if isinstance(value, list) and len(value) == 2:
                return {field: {"$gte": value[0], "$lte": value[1]}}
            raise ValueError("BETWEEN operator requires a list of two values.")

        # Handle null checks
        elif operator == Operator.IS_NULL:
            return {field: {"$eq": None}}
        elif operator == Operator.IS_NOT_NULL:
            return {field: {"$ne": None}}


from pydantic import BaseModel, ConfigDict
from datetime import datetime
from enum import Enum
from typing import Optional,Union,List,Any,Literal

class Operator(str, Enum):
    EQ = "eq"
    NE = "ne"
    GT = "gt"
    GTE = "gte"
    LT = "lt"
    LTE = "lte"
    IN = "in"
    NIN = "nin"
    LIKE = "like"
    ILIKE = "ilike"
    BETWEEN = "between"
    IS_NULL = "is_null"
    IS_NOT_NULL = "is_not_null"

class LogicalOperator(str, Enum):
    AND = "and"
    OR = "or"

class SortDirection(str, Enum):
    ASC = "ASC"
    DESC = "DESC"

class FieldTypeEnum(str, Enum):
    STR = "str"
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    LIST = "list"
    DICT = "dict"
    DATETIME = "datetime"


class BaseSchema(BaseModel):
    id:str
    uuid:str
    created_at:datetime
    updated_at:datetime

    model_config = ConfigDict(from_attributes=True)

class BaseDraftSchema(BaseSchema):
    is_draft:bool=True

class FindUniqueByFieldInput(BaseModel):
    field_name: str
    ordering: Optional[Literal["asc", "desc"]]=None
    page: Optional[int]=1
    page_size: Optional[int]=10
    search: Optional[str]=None


class FieldOperatorCondition(BaseModel):
    field: str
    operator: Operator
    value: Union[str,int,float,bool,list,dict,datetime]

class LogicalCondition(BaseModel):
    operator: LogicalOperator
    conditions: List["ConditionType"]

ConditionType = Union["LogicalCondition", "FieldOperatorCondition"]



# Top-level filter schema
class FilterSchema(BaseModel):
    operator: LogicalOperator
    conditions: List[ConditionType]

class SortSchema(BaseModel):
    field:str
    direction:SortDirection

class ListFilter(BaseModel):
    filters: Optional[FilterSchema] = None
    sort_order: Optional[List[SortSchema]] = None
    page: Optional[int] = 1
    page_size: Optional[int] = 20
    search: Optional[str] = None
    searchable_fields: Optional[List[str]] = None

class SearchOptions(BaseModel):
    search: Optional[str] = None
    sort_order: Optional[List[SortSchema]] = None
    page: Optional[int] = None
    page_size: Optional[int] = None
    total_pages: Optional[int] = None
    total_count: Optional[int] = None

class ListResponse(BaseModel):
    founds: List[Any]
    search_options: SearchOptions



class AddColumnField(BaseModel):
    column_field:str
    column_type:FieldTypeEnum
    column_default:Optional[Any] = None


from typing import Type, TypeVar, List, Dict, Any
from ..document import BaseDocument, BaseDraftDocument
from ..exceptions import BadRequestError, NotFoundError, MongoError
from pydantic import BaseModel
from pymongo import ASCENDING, DESCENDING
from beanie.operators import Set
from pymongo.errors import PyMongoError
from datetime import datetime


from ..schema import ListFilter, SortDirection, LogicalOperator, Operator, FindUniqueByFieldInput,AddColumnField, FieldTypeEnum
from ..util.operator_mappings import logical_operator_map, apply_condition

T = TypeVar("T", bound=BaseModel)

class BaseRepository:
    """
    Base repository class for doing all the database operations using Beanie for NoSQL database
    """

    def __init__(self, document:Type[BaseDocument|BaseDraftDocument]):
        self.document = document

    async def create(self,data:T):
        """
        Create a new document in the collection
        """
        obj_dict = data.__dict__ if hasattr(data, '__dict__') else dict(data)

        model_instance = self.document(**obj_dict)
        await model_instance.insert()  # Inserting the object to the DB
        return model_instance
    
    async def bulk_create(self,data:List[T]):
        """
        Create multiple documents in the collection
        """
        
        get_obj = lambda obj: obj.__dict__ if hasattr(obj, '__dict__') else dict(obj)
        model_instances = [self.document(**get_obj(item)) for item in data]
        await self.document.insert_many(model_instances)
        return model_instances
    

    async def update(self,id:str,obj:T):
        """
        Update a document by id
        """
        # Only include fields that are not None (assuming None means "not provided")
        obj_dict = {k: v for k, v in obj.__dict__.items() if v is not None}

        result = await self.document.get(id)
        if result is None:
            raise NotFoundError(detail="You're trying to update a non-existing object")

        await result.update(Set(obj_dict))

        return await self.document.get(id)

    async def get_by_attr(self,attr:str, value:any)->Type[BaseDocument]:
        """
        Get a document by a specific attribute
        """
        try:
            if hasattr(self.document,attr):
                return await self.document.find_one(getattr(self.document,attr)==value)
            else:
                raise BadRequestError(f"Attribute {attr} not found in document {self.document.__name__}")

        except Exception as e:
            raise e
        
    async def get_all(self,find:ListFilter):
        """
        Get all documents from the collection with advanced filtering, sorting, searching and pagination
        """
        # Fetching the data from MongoDB
        page = find.page or 1
        page_size = find.page_size or 20
        skip = (page - 1) * page_size

        mongo_filter = None

        # Build the query with filter
        if find.filters:
            filter_dict = find.filters.model_dump()
            mongo_filter = self.build_mongo_filter(filter_dict)
            if find.search and find.searchable_fields and len(find.searchable_fields) > 0:
                search_conditions = [
                    {field: {"$regex": f".*{find.search}.*", "$options": "i"}}
                    for field in find.searchable_fields
                ]
                search_filter = {"$or": search_conditions}
                if mongo_filter:
                    mongo_filter = {"$and": [mongo_filter, search_filter]}
                else:
                    mongo_filter = search_filter
            query = self.document.find(mongo_filter)
        else:
            query = self.document.find_all()

        count = await query.count()

        # Apply pagination
        query = query.skip(skip).limit(page_size)

        # Apply sorting if provided
        if find.sort_order and len(find.sort_order) > 0:
            order = self.get_sort_order(find.sort_order)
            query = query.sort(order)

        # Execute the query and fetch data
        data = await query.to_list()

        # Ensure data is not empty
        data = data or []


        # Calculate the total number of pages
        total_pages = (count + find.page_size - 1) // find.page_size

        # Return the data and search options (pagination details)
        response = {
            "founds": data, 
            "search_options": {
                "total_pages": total_pages,
                "total_count": count,
                "page": find.page,
                "page_size": find.page_size,
                "search": find.search,
                "sort_order": find.sort_order
            }
        }
        return response

    def get_sort_order(self,sort_order:List):
        order = []
        for sort in sort_order:
            if self.document.model_fields.get(sort.field):
                if sort.direction == SortDirection.ASC:
                    order.append((sort.field,ASCENDING))
                else:
                    order.append((sort.field,DESCENDING))
        return order
    
    def build_mongo_filter(self, filter_dict: Dict,operator:LogicalOperator=LogicalOperator.AND) -> Dict[str, Any]:
        if not isinstance(filter_dict, dict):
            return {}

        if "operator" in filter_dict and "conditions" in filter_dict:
            mongo_sub_filters = []

            for cond in filter_dict["conditions"]:
                sub_filter = self.build_mongo_filter(cond)
                if sub_filter:
                    mongo_sub_filters.append(sub_filter)

            return {
                logical_operator_map[filter_dict["operator"]]: mongo_sub_filters
            }

        elif "field" in filter_dict and "operator" in filter_dict and "value" in filter_dict:
            try:
                comp_operator = Operator(filter_dict["operator"].lower())
                return apply_condition(self.document,comp_operator,filter_dict["field"], filter_dict["value"])
            except ValueError:
                raise BadRequestError(f"Invalid comparison operator: {filter_dict['operator']}")
        else:
            raise BadRequestError(f"Invalid filter structure: {filter_dict}")
        
    async def delete(self, id: str) -> bool:
        result = await self.document.get(id)
        if result:
            await result.delete()
            return True
        else:
            raise NotFoundError("Youre trying to delete a non-existing object")
        
    async def get_unique_values(self, schema: FindUniqueByFieldInput):
        try:

            field_name = schema.field_name
            ordering = schema.ordering or "asc"
            page = schema.page or 1
            page_size = schema.page_size or 10
            search = schema.search or None

            if not field_name:
                raise ValueError("field_name is required")

            # Build the aggregation pipeline
            pipeline = []

            # Apply search 
            if search:
                pipeline.append({
                    "$match": {
                        field_name: {"$regex": f".*{search}.*", "$options": "i"}
                    }
                })
            # Group by field for distinct values
            pipeline.append({
                "$group": {
                    "_id": f"${field_name}"
                }
            })
            if schema.ordering:
                pipeline.append({
                    "$sort": {
                        "_id": 1 if schema.ordering == "asc" else -1
                    }
                })


            # Count total
            count_pipeline = pipeline + [{"$count": "total"}]
            count_result = await self.document.get_motor_collection().aggregate(count_pipeline).to_list(length=1)
            total_count = count_result[0]["total"] if count_result else 0

            # Add pagination
            skip = (page - 1) * page_size
            pipeline.append({"$skip": skip})
            pipeline.append({"$limit": page_size})

            # Execute aggregation to get the distinct values
            results = await self.document.get_motor_collection().aggregate(pipeline).to_list(length=None)
            values = [r["_id"] for r in results]

            return {
                    "founds": values,
                    "search_options": {
                        "page": page,
                        "page_size": page_size,
                        "ordering": ordering,
                        "total_count": total_count
                    }
            }

        except PyMongoError as e:
            raise MongoError(detail=str(e))
        except Exception as e:
            raise BadRequestError(detail=str(e))

    async def add_field(self,column:AddColumnField):
        collection = self.document.get_motor_collection()

        try:

            result = await collection.update_many(
                {column.column_field: {"$exists": False}},
                {"$set": {column.column_field: self.get_default_for_type(column.column_type)}},
                hint="_id_"
            )

            return True
        except Exception as e:
            raise MongoError(detail=str(e))

    def get_default_for_type(self, field_type: FieldTypeEnum) -> Any:
        default_values = {
            FieldTypeEnum.STRING: "",
            FieldTypeEnum.INTEGER: 0,
            FieldTypeEnum.BOOLEAN: False,
            FieldTypeEnum.FLOAT: 0.0,
            FieldTypeEnum.LIST: [],
            FieldTypeEnum.DICT: {},
            FieldTypeEnum.DATE: datetime.now()
        }
        return default_values.get(field_type, None)

